/*******************************************************************************
 * Copyright (c) 2007 Supelec.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This file is part of the ADLV tool
 * Version 1.0 (December 5, 2007)
 *
 * Contributors:
 *     Computer Science Department, Supelec : initial implementation
 * Contacts:
 *     Dominique.Marcadet@supelec.fr
 *     Frederic.Boulanger@supelec.fr
 *     Christophe.Jacquet@supelec.fr
 *******************************************************************************/
package fr.supelec.adlv.checker;

import java.io.Serializable;
import java.util.Collection;
import java.util.Iterator;

public class Util {
	/**
	 * Transforms a collection of objects into a string of comma-separated
	 * values. The values are obtained by calling toString() on each member
	 * of the collection.
	 *
	 * @param c a collection
	 * @return a comma-separated string
	 */
	public static <A> String collectionToString(Collection<A> c, String sep, StringConverter<A> maker) {
		StringBuffer res = new StringBuffer();
		boolean notFirst = false;
		for(A o : c) {
			if(notFirst) res.append(sep);
			res.append(maker.convert(o));
			notFirst = true;
		}
		return res.toString();
	}


	public static <A> String collectionToString(Collection<A> c) {
		return collectionToString(c, ", ", new StringConverter<A> () {
			public String convert(A a) {
				return a.toString();
			}

		});
	}


	/**
	 * This class represents pairs of objects, possibly of different types.
	 *
	 * @param <A> type of the first element
	 * @param <B> type of the second element
	 */
	public static class Pair<A, B> implements Serializable {
		private static final long serialVersionUID = 4775196068161311926L;

		private final A a;
		private final B b;

		/**
		 * Constructs a new pair.
		 *
		 * @param a the first element
		 * @param b the second element
		 */
		public Pair(A a, B b) {
			this.a = a;
			this.b = b;
		}

		/**
		 * Returns the first element of the pair.
		 * @return the first element
		 */
		public A first() {
			return a;
		}

		/**
		 * Returns the second element of the pair.
		 * @return the second element
		 */
		public B second() {
			return b;
		}
	}

	/**
	 * Returns ONE element from a given collection.
	 *
	 * This method is intended to be called on singletons, thus returning the
	 * only element of the singleton.
	 *
	 * @param <A> any type
	 * @param c a collection of objects of type {@code <A>}
	 * @return one element from {@code c}, or {@code null} if {@code c} is
	 * empty
	 */
	public static <A> A pickFromSingleton(Collection<A> c) {
		for(A item : c) {
			return item;
		}
		return null;
	}


	/**
	 * Returns an Iterable object returning a given iterator. This is a
	 * utility function used to iterate iterator with the generalized
	 * for( : ) syntax.
	 *
	 * @param <C> type of the elements
	 * @param it iterator
	 * @return an iterable
	 */
	public static <C> Iterable<C> makeIterable(final Iterator<C> it) {
		return new Iterable<C>() {
			public Iterator<C> iterator() {
				return it;
			}

		};
	}

	public static interface StringConverter<A> {
		public String convert(A a);
	}
}
