/*******************************************************************************
 * Copyright (c) 2007 Supelec.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This file is part of the ADLV tool
 * Version 1.0 (December 5, 2007)
 *
 * Contributors:
 *     Computer Science Department, Supelec : initial implementation
 * Contacts:
 *     Dominique.Marcadet@supelec.fr
 *     Frederic.Boulanger@supelec.fr
 *     Christophe.Jacquet@supelec.fr
 *******************************************************************************/
package fr.supelec.adlv.checker.logic;

import java.util.List;
import java.util.Set;

import fr.supelec.adlv.Connection;
import fr.supelec.adlv.checker.PortInstance;

public abstract class Expr implements Comparable<Expr> {
	/**
	 * Accepts a visitor.
	 * @param visitor the visitor
	 */
	public abstract void accept(ExprVisitor visitor);


	protected abstract Expr develop(boolean cnf);
	public abstract Expr simplify();

	/**
	 * Returns a new expression, composed uniquely of NaryOrs, NaryAnds, and
	 * Predicates. Two equivalent formulae should yield the same canonical
	 * form (unless there is a bug in this method!)
	 * @return the canonical form
	 */
	public Expr canonical(boolean cnf) {
		// fixed point
		Expr e = null;
		Expr newE = this;
		do {
			e = newE;
			newE = e.develop(cnf);  // build the DNF
			///System.out.println("  %% " + newE);
		} while(!newE.equals(e));

		if(e instanceof Or) {
			e = ((Or)e).flatten();
		} else if(e instanceof And) {
			e = ((And)e).flatten();
		}

		return e.simplify();
	}

	public Relation compare(Expr e) {
		Relation res = Relation.NONE;

		if(isTautology(new Or(new Not(this), e)))
			res = Relation.IMPLIES;

		if(isTautology(new Or(new Not(e), this))) {
			if(res == Relation.IMPLIES) res = Relation.EQUIV;
			else res = Relation.IS_IMPLIED;
		}

		return res;
	}

	private static boolean isTautology(Expr e) {
		Expr cano = e.canonical(true);
		//System.out.print("       %%% isTautology: " + e + "  cano  " + cano + "--> ");
		if(cano instanceof Constant) {
			Constant c = (Constant)cano;
			//System.out.println(c.getValue());
			return c.getValue();
		}
		//System.out.println(false);
		return false;
	}


	@Override
	public int hashCode() {
		return toString().hashCode();
	}

	public int compareTo(Expr e) {
		return toString().compareTo(e.toString());
	}

	@Override
	public boolean equals(Object o) {
		if(o instanceof Expr) {
			Expr e = (Expr)o;
			return toString().equals(e.toString());
		}
		return false;
	}


	/**
	 * Tells whether an expression is a <em>temporal expression</em>. A
	 * <em>temporal expression</em> is defined as an expression containing
	 * at least one temporal operator.
	 *
	 * @return {@code true} if the expression is a temporal expression,
	 * {@code false} otherwise
	 */
	public abstract boolean isTemporal();


	@Override
	public String toString() {
		throw new RuntimeException("All Expr concrete subclasses MUST override toString()");
	}

	public abstract Set<PortInstance> getInputSignals();

	public static final Expr TRUE = new Constant(true);
	public static final Expr FALSE = new Constant(false);

	public abstract ExprBundle getBundle();

	public static enum Relation {
		IMPLIES("Implies"), IS_IMPLIED("IsImplied"), EQUIV("Equiv"), NONE("None");

		private final String str;

		private Relation(String str) {
			this.str = str;
		}

		@Override
		public String toString() {
			return str;
		}
	}

	public boolean isBundled() {
		return getBundle().isBundled;
	}

	public static class ExprBundle {
		public final Expr inner, outer;
		public final boolean isBundled;

		public ExprBundle(Expr inner, Expr outer, boolean isBundled) {
			this.inner = inner;
			this.outer = outer;
			this.isBundled = isBundled;
		}
	}

	/**
	 * Tells whether two port are connected.
	 *
	 * @param p1 first port
	 * @param p2 second port
	 * @return {@code true} if the ports are connected, {@code false} otherwise
	 */
	protected static boolean areConnectedOrEqual(PortInstance p1, PortInstance p2) {
		if(p1 == p2) return true;

		// 1) p1 -> p2
		List<Connection> conns = p1.getPort().getConnectionFrom();
		for(Connection c : conns)
			if(
					c.getToPort() == p2.getPort() &&
					c.getToComponent() == p2.getComponent() &&
					c.getFromComponent() == p1.getComponent()
			) return true;

		// 2) p1 <- p2
		conns = p1.getPort().getConnectionTo();
		for(Connection c : conns)
			if(
					c.getFromPort() == p2.getPort() &&
					c.getFromComponent() == p2.getComponent() &&
					c.getToComponent() == p1.getComponent()
			) return true;

		return false;
	}
}
