/*******************************************************************************
 * Copyright (c) 2007 Supelec.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This file is part of the ADLV tool
 * Version 1.0 (December 5, 2007)
 *
 * Contributors:
 *     Computer Science Department, Supelec : initial implementation
 * Contacts:
 *     Dominique.Marcadet@supelec.fr
 *     Frederic.Boulanger@supelec.fr
 *     Christophe.Jacquet@supelec.fr
 *******************************************************************************/
package fr.supelec.adlv.checker.logic;

import java.util.Iterator;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import fr.supelec.adlv.checker.PortInstance;
import fr.supelec.adlv.checker.Util;

public abstract class NaryExpr extends Expr {
	// "operands" being a SortedSet guarantees the uniqueness of canonical
	// formulae
	protected SortedSet<Expr> operands = new TreeSet<Expr>();

	@Override
	protected Expr develop(boolean cnf) {
		assert false : "develop() should not be called here.";
		return null;
	}

	public void add(Expr e) {
		operands.add(e);
	}

	public void append(NaryExpr ne) {
		operands.addAll(ne.operands);
	}

	public String toString(String op) {
		StringBuffer res = new StringBuffer("[ ");
		int i = 0;
		for(Expr e : operands) {
			if(i>0) res.append(' ').append(op).append(i).append(' ');
			i++;
			res.append(e.toString());
		}
		res.append(" ]");
		return res.toString();
	}

	@Override
	public void accept(ExprVisitor visitor) {
		System.err.println("Warning: accept() should not be called on NaryExpr.");
	}

	@Override
	public Set<PortInstance> getInputSignals() {
		return null; // not implemented for this class
	}


	/**
	 * Tells whether the expression contains a predicate {@code p} and its
	 * negation {@code p.neg()}.
	 *
	 * @return {@code true} if so, {@code false} otherwise
	 */
	protected boolean containsInversePredicates() {
		for(Predicate p1 : iterableOverPredicates(null))
			for(Predicate p2 : iterableOverPredicates(p1))
				if(p1.isNegationOf(p2)) return true;

		return false;
	}

	private Iterator<Predicate> iterateOverPredicates(final Predicate startAfter) {
		return new Iterator<Predicate>() {
			private boolean hasNext;
			private Predicate next;
			private final Iterator<Expr> it;

			// constructor
			{
				it = operands.iterator();

				// if need be, move the cursor at position {@code startAfter}
				if(startAfter != null)
					while(it.hasNext())
						if(it.next() == startAfter) break;

				step();
			}

			private void step() {
				Expr nextExpr = null;
				do {
					hasNext = it.hasNext();
					if(it.hasNext()) nextExpr = it.next();
				} while(hasNext && !(nextExpr instanceof Predicate));

				if(hasNext)
					next = (Predicate)nextExpr;
			}

			public boolean hasNext() {
				return hasNext;
			}

			public Predicate next() {
				Predicate res = next;
				step();
				return res;
			}

			public void remove() {
				// not implemented
			}

		};
	}

	private Iterable<Predicate> iterableOverPredicates(final Predicate startAfter) {
		return Util.makeIterable(iterateOverPredicates(startAfter));
	}

	@Override
	public boolean isTemporal() {
		throw new RuntimeException("isTemporal should not be called on NaryExpr.");
	}

	@Override
	public ExprBundle getBundle() {
		throw new RuntimeException("getBundle() should not be called on NaryExpr.");
	}
}
