/*******************************************************************************
 * Copyright (c) 2007 Supelec.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This file is part of the ADLV tool
 * Version 1.0 (December 5, 2007)
 *
 * Contributors:
 *     Computer Science Department, Supelec : initial implementation
 * Contacts:
 *     Dominique.Marcadet@supelec.fr
 *     Frederic.Boulanger@supelec.fr
 *     Christophe.Jacquet@supelec.fr
 *******************************************************************************/
package fr.supelec.adlv.generator;

import java.io.File;
import java.io.IOException;
import java.util.Collection;
import java.util.HashSet;

import fr.supelec.adlv.*;

import fr.supelec.adlv.parser.EmfAdlvHelper;

public class Idl3Generator extends AdlvGenerator {

	Idl3Generator( EmfAdlvHelper helper ) {
		helper_ = helper;
	}
	
	void generate( String resultsPath ) throws IOException {
		
		for( Object o : helper_.getApplication().getComponents() ) {
			Component c = (Component) o;
			ComponentType t = c.getType();
			generateIdl3FileForComponent( t, resultsPath );
		}
		
		generateIdl3FileForDynamicRouting( resultsPath );
		generateIdl3FileForEnvironment( resultsPath );
	}
	
	private void generateIdl3FileCommonPart( String cname, String mname, String events, String ports, boolean simulation ) throws IOException {
		writeCppHeaderPrelude( mname );
		
		out_.write( "\n" );
		out_.write( "#include \"" + STREAM_TYPE_INTERFACE_FILENAME
				  + "." + IDL3_EXTENSION + "\"\n" );
		out_.write( events + "\n" );

		out_.write( "\n" );
		out_.write( "//module " + mname + " {\n" );
		out_.write( "\n" );

		out_.write( "\n" );
		out_.write( "component " + cname + " {\n" );
		out_.write( "\n" );
		
		out_.write( ports + "\n" );
		
		if( simulation && STATIC_DEPLOYMENT ) {
			out_.write( "    // Simulation component\n" );
			out_.write( "    provides SimulationComponent start;" );
			out_.write( "\n" );
		}
		
		out_.write( "};\n" );
		out_.write( "\n" );

		out_.write( "\n" );
		out_.write( "home " + cname + "Home manages " + cname + " {\n" );
		out_.write( "    //factory new( in string name );\n" );
		out_.write( "};\n" );
		out_.write( "\n" );

		out_.write( "\n" );
		out_.write( "//};\n" );
		out_.write( "\n" );

		writeCppHeaderPostlude();
	}
	
	private String generateIdl3FileCommonPartEvents( Collection< Port > ports ) throws IOException {
		StringBuffer events_str = new StringBuffer();
		HashSet< String > eventsModule = new HashSet< String >();
		for( Port p : ports ) {
			TokenType t = p.getTokenType();
			if( t instanceof EventType ) {
				eventsModule.add( t.getModule().getName() );
			}
		}
		for( String s : eventsModule ) {
			events_str.append( "#include \"" + s + EVENT_TYPE_FILE_SUFFIX
					         + "." + IDL3_EXTENSION + "\"\n" );
		}
		
		return events_str.toString();
	}

	private String generateIdl3FileCommonPartStream( Collection< Port > ports, boolean simulation ) throws IOException {
		StringBuffer ports_str = new StringBuffer();
		for( Port p : ports ) {
			TokenType t = p.getTokenType();
			if( t instanceof StreamType ) {
				// StreamType
				StreamType s = (StreamType) t;
				if(( simulation && ( p instanceof ConsumerPort ) ) || ( ! simulation && ( p instanceof ProducerPort ))) {
					// source
					ports_str.append( "    // source " + s.getName() + " "
							        + p.getName() + ";\n" );
					//ports_str.append( "    uses " + STI_MODULE_NAME + "::"
					ports_str.append( "    uses "
							        + getStreamType( s.getType().getName() ) + " "
							        + p.getName() + ";\n" );
					ports_str.append( "\n" );
				}
				else {
					// sink
					ports_str.append( "    // sink " + s.getName() + " "
							        + p.getName() + ";\n" );
					//ports_str.append( "    provides " + STI_MODULE_NAME + "::"
					ports_str.append( "    provides "
							        + getStreamType( s.getType().getName() ) + " "
							        + p.getName() + ";\n" );
					ports_str.append( "\n" );
				}
			}
			else {
				// EventType
				if(( simulation && ( p instanceof ConsumerPort ) ) || ( ! simulation && ( p instanceof ProducerPort ))) {
					// publishes
					ports_str.append( "    // publishes " + t.getName() + " "
							        + p.getName() + ";\n" );
					ports_str.append( "    publishes " + t.getModule().getName()
							        + "::" + t.getName() + " "
							        + p.getName() + ";\n" );
					ports_str.append( "\n" );
				}
				else {
					// consumes
					ports_str.append( "    // consumes " + t.getName() + " "
							        + p.getName() + ";\n" );
					ports_str.append( "    consumes " + t.getModule().getName()
							        + "::" + t.getName() + " "
							        + p.getName() + ";\n" );
					ports_str.append( "\n" );
				}
			}
		}
		
		return ports_str.toString();
	}

	private void generateIdl3FileCommonPart( String cname, String mname, Collection< Port > ports, boolean simulation ) throws IOException {
		String events_str = generateIdl3FileCommonPartEvents( ports );
		String ports_str = generateIdl3FileCommonPartStream( ports, simulation );
		generateIdl3FileCommonPart( cname, mname, events_str, ports_str, simulation );
	}
	
	private void generateIdl3FileForComponent( ComponentType component, String resultsPath ) throws IOException {
		String cname = component.getName();
		String componentDirectory = resultsPath + "/" + cname;
		( new File( componentDirectory ) ).mkdir();

		openFile( componentDirectory, cname, IDL3_EXTENSION );
		
		generateIdl3FileCommonPart( cname, component.getModule().getName(), component.getPorts(), false );
		
		closeFile();
	}
	
	private void generateIdl3FileForEnvironment( String resultsPath ) throws IOException {
		String componentDirectory = resultsPath + "/" + ENVIRONMENT_NAME;
		( new File( componentDirectory ) ).mkdir();

		openFile( componentDirectory, ENVIRONMENT_NAME, IDL3_EXTENSION );
		
		HashSet< Port > ports = new HashSet< Port >();
		ports.addAll( helper_.getApplicationExternalConsumerPorts() );
		ports.addAll( helper_.getApplicationExternalProducerPorts() );
		String events_str = generateIdl3FileCommonPartEvents( ports );
		String ports_str = generateIdl3FileCommonPartStream( ports, true );
		
		// We also provide the internal consumer ports (those used for dynamic
		// configuration) so that they can be examined
		ports.clear();
		ports.addAll( helper_.getApplicationInternalConsumerPorts() );
		String ports_str2 = generateIdl3FileCommonPartStream( ports, false );

		Application application = helper_.getApplication();
		generateIdl3FileCommonPart( ENVIRONMENT_NAME, application.getModule().getName(), events_str, ports_str + ports_str2, true );
		
		closeFile();
	}
	
	private void generateIdl3FileForDynamicRouting( String resultsPath ) throws IOException {
		String componentDirectory = resultsPath + "/" + DYNAMIC_ROUTING_NAME;
		( new File( componentDirectory ) ).mkdir();

		openFile( componentDirectory, DYNAMIC_ROUTING_NAME, IDL3_EXTENSION );
		
		Application application = helper_.getApplication();
		
		HashSet< String > eventsModule = new HashSet< String >();
		HashSet< Port > usedPorts = new HashSet< Port >();
		for( Port p : application.getPorts() ) {
			// Looking for event consumers
			if( p instanceof ProducerPort ) continue;
			TokenType t = p.getTokenType();
			if( t instanceof StreamType ) continue;
			
			// that are used for dynamic configuration
			if((( ConsumerPort ) p ).getSteps().isEmpty() ) continue;
			
			eventsModule.add( t.getModule().getName() );
			usedPorts.add( p );
			
			for( ConfigurationStep step : (( ConsumerPort ) p ).getSteps() ) {
				if( step instanceof ComponentConfigurationStep ) continue;
				
				usedPorts.add((( ConnectionConfigurationStep ) step ).getConnection().getFromPort() );
				usedPorts.add((( ConnectionConfigurationStep ) step ).getConnection().getToPort() );
			}
		}
		
		StringBuffer events_str = new StringBuffer();
		for( String s : eventsModule ) {
			events_str.append( "#include \"" + s + EVENT_TYPE_FILE_SUFFIX
					      + "." + IDL3_EXTENSION + "\"\n" );
		}

		StringBuffer ports_str = new StringBuffer();
		for( Port p : usedPorts ) {
			boolean applicationPort = p.getApplication() != null;
			TokenType t = p.getTokenType();
			if( t instanceof StreamType ) {
				StreamType s = (StreamType) t;
				if( p instanceof ProducerPort ) {
					ports_str.append( "    "
							        + ( applicationPort ? "uses " : "provides " )
      							    //+ STI_MODULE_NAME + "::"
		      					    + getStreamType( s.getType().getName() ) + " "
		      				        + p.getName() + ";\n" );
				}
				else {
					ports_str.append( "    "
						            + ( applicationPort ? "provides " : "uses " )
							        //+ STI_MODULE_NAME + "::"
							        + getStreamType( s.getType().getName() ) + " "
							        + p.getName() + ";\n" );
				}
			}
			else {
				// EventType
				if( p instanceof ProducerPort ) {
					ports_str.append( "    "
							        + ( applicationPort ? "publishes " : "consumes " )
							        + t.getModule().getName()
							        + "::" + t.getName() + " "
							        + p.getName() + ";\n" );
				}
				else {
					ports_str.append( "    "
							  + ( applicationPort ? "consumes " : "publishes " )
							  + t.getModule().getName()
							  + "::" + t.getName() + " "
							  + p.getName() + ";\n" );
				}
			}
			ports_str.append( "\n" );
		}

		generateIdl3FileCommonPart( DYNAMIC_ROUTING_NAME, application.getModule().getName(), events_str.toString(), ports_str.toString(), false );
		
		closeFile();
	}
	
	private EmfAdlvHelper helper_;
}
