header {
/*******************************************************************************
 * Copyright (c) 2007 Supelec.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * This file is part of the ADLV tool
 * Version 1.0 (December 5, 2007)
 *
 * Contributors:
 *     Computer Science Department, Supelec : initial implementation
 * Contacts:
 *     Dominique.Marcadet@supelec.fr
 *     Frederic.Boulanger@supelec.fr
 *     Christophe.Jacquet@supelec.fr
 *******************************************************************************/

package fr.supelec.adlv.parser;

import java.util.Vector;
import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;

import org.eclipse.emf.common.util.EList;

import fr.supelec.adlv.*;

}

/**
 * Part of the grammar is based on IDL3.g :
 *
 *
 *  This is a complete parser for the IDL language as defined
 *  by the CORBA 3.0.2 specification.  It will allow those who
 *  need an IDL parser to get up-and-running very quickly.
 *  Though IDL's syntax is very similar to C++, it is also
 *  much simpler, due in large part to the fact that it is
 *  a declarative-only language.
 *
 *  Some things that are not included are: Symbol table construction
 *  (it is not necessary for parsing, btw) and preprocessing (for
 *  IDL compiler #pragma directives). You can use just about any
 *  C or C++ preprocessor, but there is an interesting semantic
 *  issue if you are going to generate code: In C, #include is
 *  a literal include, in IDL, #include is more like Java's import:
 *  It adds definitions to the scope of the parse, but included
 *  definitions are not generated.
 *
 *  Jim Coker, jcoker@magelang.com
 *  Gary Duzan, gduzan@bbn.com
 */

class ADLVParser extends Parser;
options {
    exportVocab = ADLV;
    buildAST = true;
    ASTLabelType = "NumberedAST";
    k = 4;
}

{
    /**
     * The helper used to find in the model
     */
    private EmfAdlvHelper helper_ = new EmfAdlvHelper();
    
    public EmfAdlvHelper getHelper() {
        return helper_;
    }
    
    /**
     * The factory used to instantiate the model
     */
    private AdlvFactory factory_ = AdlvFactory.eINSTANCE;

    /**
     * The application
     */
    private Application application_ = null;
    
    /**
     * The current module name
     */
    private String current_module_name_ = null;
    
    /**
     * The Component that we are parsing
     */
    private ComponentType component_ = null;
    
    /**
     * There should be a way, in EMF, to get a map automatically generated
     * (see FAQ "How do I create a Map in EMF?"). But I don't know how to do this
     * with Papyrus
     */
    /* Gestion des erreurs */
    private Vector< String > errors = new Vector< String >();
    private Vector< String > warnings = new Vector< String >();

    /**@return The number of errors found in the parser*/
    public int errorNumber(){
        return errors.size();
    }

    /**@return The number of warnings found in the parser*/
    public int warningNumber(){
        return warnings.size();
    }

    /**@return All errors messages found in the parser*/
    public String [] errors () {
        String [] ret = new String [errors.size()];
        for (int i = 0; i < ret.length; ++i)
            ret[i] = (String)errors.elementAt(i);
        return ret;
    }

    /**@return All warning messages found in the parser*/
    public String [] warnings () {
        String [] ret = new String [warnings.size()];
        for (int i = 0; i < ret.length; ++i)
            ret[i] = (String)warnings.elementAt(i);
        return ret;
    }

    public void reportError(RecognitionException ex) {
        errors.addElement(ex.toString());
    }

    public void reportError(String s) {
        errors.addElement(s);
    }

    public void reportError(int line, int col, String s) {
    	reportError(line + ":" + col + ": " + s);
    }

    public void reportWarning(String s) {
        warnings.addElement(s);
    }
}

// debut des regles de la grammaire

specification
    :   module_definition SEMI!
    ;

module_definition
    :   "module"^
        i: identifier
        {
            current_module_name_ = #i.getText();
            Module module = factory_.createModule();
            module.setName( current_module_name_ );
            helper_.addModule( module );
        }
        LCURLY! definition_list RCURLY!
    ;

definition_list
    :   ( definition )+
        application SEMI!
    ;


definition
    :   event_type SEMI!
    |   stream_type SEMI!
    |   component_type SEMI!
    ;

event_type
    :   "eventtype"^
        i: identifier
        {
            EventType e = factory_.createEventType();
            e.setName( #i.getText() );
            e.setModule( helper_.getModule( current_module_name_ ));
        }
        LCURLY!  RCURLY!
    ;

stream_type
    {
        DataType dt = null;
    }
    :    "streamtype"^
        LT! dt = type_spec GT!
        i: identifier
        {
            StreamType s = factory_.createStreamType();
            s.setName( #i.getText() );
            s.setModule( helper_.getModule( current_module_name_ ));
            s.setType( dt );
        }
    ;
type_spec returns [DataType dt = null]
    :    "boolean" { dt = DataType.BOOLEAN; }
    |    "long"     { dt = DataType.LONG; }
    |    "double"  { dt = DataType.DOUBLE; }
    ;

component_type
    :    control_component_type
    |    processing_component_type
    |    internal_component_type
    ;

control_component_type
    :   "control" "component"^
        i: identifier
        {
            component_ = factory_.createControlComponentType();
            component_.setName( #i.getText() );
            component_.setModule( helper_.getModule( current_module_name_ ));
        }
        LCURLY! control_component_body RCURLY!
        {
            component_ = null;
        }
    ;

processing_component_type
    :   "processing" "component"^
        i: identifier
        {
            component_ = factory_.createProcessingComponentType();
            component_.setName( #i.getText() );
            component_.setModule( helper_.getModule( current_module_name_ ));
        }
        LCURLY! processing_component_body RCURLY!
        {
            component_ = null;
        }
    ;

internal_component_type
    :   "internal" "component"^
        i: identifier
        {
            component_ = factory_.createInternalComponentType();
            component_.setName( #i.getText() );
            component_.setModule( helper_.getModule( current_module_name_ ));
        }
        LCURLY! internal_component_body RCURLY!
        {
            component_ = null;
        }
    ;

control_component_body
    {
        String tool = null;
        Port p = null;
    }
    :      tool = tool_dcl { component_.setTool( tool ); }
           (
               (
                       p = consumes_dcl    { component_.getPorts().add( p ); }
                   |   p = publishes_dcl   { component_.getPorts().add( p ); }
                )
                SEMI!
           )+
    ;

processing_component_body
    {
        String tool = null;
        Port p = null;
    }
    :  tool = tool_dcl { component_.setTool( tool ); }
       (
               (
                       p = source_dcl { component_.getPorts().add( p ); }
                   |   p = sink_dcl   { component_.getPorts().add( p ); }
               )
               SEMI!
           )+
    ;

internal_component_body
    {
        Port p = null;
        Expression e = null;
        DataType dt = null;
    }
    :  (    p = source_dcl         { component_.getPorts().add( p ); }
            (
                    SEMI!
                |     source_def[p]
            )
        |    p = sink_dcl     SEMI!  { component_.getPorts().add( p ); } /* sink_def : not used */
        |    p = consumes_dcl        { component_.getPorts().add( p ); }
            (
                   SEMI!
                   // There are 2 consumes_def : 1 for internal_component and 1 for application
                |  component_consumes_def[p]
            ) 
        |     p = publishes_dcl       { component_.getPorts().add( p ); }
            (
                    SEMI!
                |   e = publishes_def { (( ProducerPort ) p ).setEventEmitExpression( e ); }
            )
        |   "attribute"^
            dt = type_spec
            i: identifier
            SEMI!
            {
                Attribute a = factory_.createAttribute();
                a.setName( #i.getText() );
                a.setType( dt );
                (( InternalComponentType ) component_).getAttributes().add( a );
            }
       )+
    ;

tool_dcl returns [String tool = null]
    :   "tool"^
        i: identifier { tool = #i.getText(); }
        SEMI!
    ;

source_dcl returns [ProducerPort p = null]
    :   "source"^
        s: identifier
        i: identifier
        {
            p = factory_.createProducerPort();
            p.setName( #i.getText() );
            p.setTokenType( helper_.getStreamType( current_module_name_, #s.getText() ));
        }
    ;

source_def [Port p]
    :   LCURLY!
        "output"^
        a:identifier
        SEMI!
        {
            (( ProducerPort ) p ).setOutputAttribute( helper_.getComponentAttribute( component_, #a.getText() ));
         }
        RCURLY!
    ;

sink_dcl returns [ConsumerPort p = null]
    :   "sink"^
        s:identifier
        i:identifier
        {
            p = factory_.createConsumerPort();
            p.setName( #i.getText() );
            p.setTokenType( helper_.getStreamType( current_module_name_, #s.getText() ));
        }
    ;

sink_def
    :   LCURLY! /* ERROR */ RCURLY!
    ;

consumes_dcl returns [ConsumerPort p = null]
    :   "consumes"^
        e:identifier
        i:identifier
        {
            p = factory_.createConsumerPort();
            p.setName( #i.getText() );
            p.setTokenType( helper_.getEventType( current_module_name_, #e.getText() ));
        }
    ;

component_consumes_def [Port p]
    {
        Expression e = null;
    }
    :   LCURLY!
        a:identifier
        "="
        e = expression
        SEMI!
        {
            (( ConsumerPort ) p ).setSetAttribute( helper_.getComponentAttribute( component_, #a.getText() ));
            (( ConsumerPort ) p ).setSetExpression( e );
        }
        RCURLY!
    ;

application_consumes_def [Port p]
    {
        ConfigurationStep c = null;
    }
    :   LCURLY!
        (
            c = configuration_step SEMI!
            {
                (( ConsumerPort ) p ).getSteps().add( c );
            }
        )*
        RCURLY!
    ;

publishes_dcl returns [ProducerPort p = null]
    :   "publishes"^
        e:identifier
        i:identifier
        {
            p = factory_.createProducerPort();
            p.setName( #i.getText() );
            p.setTokenType( helper_.getEventType( current_module_name_, #e.getText() ));
        }
    ;

publishes_def returns [Expression expr = null]
    :   LCURLY!
        "when"^ expr = expression SEMI!
        RCURLY!
    ;

configuration_step returns [ConfigurationStep step = null]
    :   "activate"^
        i: identifier
        {
        	if(#i == null) {
        		reportError("syntax error. \"activate component_name\" was expected.");
        		return null;
        	}
        	String compName = #i.getText();
            Component c = helper_.getApplicationComponent( compName );
            if(c == null)
            	reportError(#i.getLine(), #i.getColumn(), "component " + compName + " not found.");
            ActivateComponentStep a = factory_.createActivateComponentStep();
            a.setComponent( c );
            step = a;
        }
    |   "deactivate"^
        j: identifier
        {
            Component c = helper_.getApplicationComponent( #j.getText() );
            DeactivateComponentStep a = factory_.createDeactivateComponentStep();
            a.setComponent( c );
            step = a;
        }
    |   step = connection_step
    ;

connection_step returns [ConnectionConfigurationStep cc = null]
    { Expression e = null; }
    :   e = binaryExpression
        {
        	if( ! ( e instanceof BinaryExpression )) {
                reportError(e + " is not a ConnectExpression" );
        	}
        	BinaryExpression b = ( BinaryExpression ) e;
        	if( ! b.getOperator().equals( BinaryOperator.CONNECTED_TO_FROM )) {
                reportError( "Error : " + e + " is not a ConnectExpression" );
        	}
        	
        	PortExpression s = ( PortExpression ) b.getRightExpression();
            PortExpression d = ( PortExpression ) b.getLeftExpression();
        	
            cc = factory_.createConnectionConfigurationStep();
            Connection c = factory_.createConnection();
            cc.setConnection( c );

            c.setFromPort( s.getPort() );
            if( s.getComponent() == null ) {
                c.setFromApplication( application_ );
            }
            else {
                c.setFromComponent( s.getComponent() );
            }
            
            c.setToPort( d.getPort() );
            if( d.getComponent() == null ) {
                c.setToApplication( application_ );
            }
            else {
                c.setToComponent( d.getComponent() );
            }
        }
    ;

application
    :   "application"^
        i: identifier
        {
            application_ = factory_.createApplication();
            application_.setName( #i.getText() );
            application_.setModule( helper_.getModule( current_module_name_ ));
            helper_.setApplication( application_ );
        }
        LCURLY!
        application_body
        RCURLY!
    ;

application_body
    {
        Port p = null;
        List<SafetyClause> c;
    }
    :   ( 
                p = consumes_dcl { application_.getPorts().add( p ); }
                (
                        SEMI!
                    |     application_consumes_def[p]
                )
            /*|     publishes_dcl : not used ? */
            |    p = sink_dcl   SEMI! { application_.getPorts().add( p ); }
            |    p = source_dcl SEMI! { application_.getPorts().add( p ); }
            |    component_use  SEMI!
        )*
        init_proc
        static_proc
        ( 
                c = never_clause   { application_.getSafetyClause().addAll(c); }
            |   c = always_clause  { application_.getSafetyClause().addAll(c); }
        )*
    ;

component_use
    :   "instance"^
        c: identifier
        i: identifier
        {
            Component ci = factory_.createComponent();
            ci.setName( #i.getText() );
            ci.setType( helper_.getComponentType( current_module_name_, #c.getText() ));
            application_.getComponents().add( ci );
        }
    ;

init_proc
    {
        ConfigurationStep cs = null;
    }
    :   "init"^
        LCURLY!
        (
            cs = configuration_step SEMI!
            {
                application_.getInitialSteps().add( cs );
            }
        )*
        RCURLY!
    ;

// TODO : init_proc and static_proc differ only by the starting keyword
static_proc
    {
        ConfigurationStep cs = null;
    }
    :   "static"^
        LCURLY!
        (
            cs = configuration_step SEMI!
            {
                application_.getStaticConfiguration().add( cs );
            }
        )*
        RCURLY!
    ;

expression returns [Expression expr = null]
    {
        Expression right = null;
        BinaryOperator op = null;
    }
    :   expr = binaryExpression 
        (
            op = logicalOperator
            right = binaryExpression
            {
                BinaryExpression e = factory_.createBinaryExpression();
                e.setLeftExpression( expr );
                e.setRightExpression( right );
                e.setOperator( op );
                expr = e;
            }
        )*
    ;

logicalOperator returns [BinaryOperator op = null]
    :    "since"     { op = BinaryOperator.SINCE; }
    |    IMPLIES!    { op = BinaryOperator.IMPLIES; }
    |    "backto"    { op = BinaryOperator.BACK_TO; }
    |    LOGICAL_AND { op = BinaryOperator.AND; }
    |    LOGICAL_OR  { op = BinaryOperator.OR; }
    ;

binaryExpression  returns [Expression expr = null]
    {
        SimpleExpression left = null, right = null;
        BinaryOperator op = null;
        boolean connect = false;
    }
    :   left = simpleExpression { expr = left; }
        (
            (
                    op = compareOperator
                |   op = connectOperator { connect = true; }
            )
            right = simpleExpression
            {
                BinaryExpression e = factory_.createBinaryExpression();
                e.setLeftExpression( left );
                e.setRightExpression( right );
                e.setOperator( op );
                expr = e;
                
                if( connect ) {
                	if( ! ( left instanceof PortExpression )) {
                		reportError( "Error : " + left + " is not a PortExpression" );
                	}
                    if( ! ( right instanceof PortExpression )) {
                        reportError( "Error : " + right + " is not a PortExpression" );
                    }
                }
            }
        )?
    ;

compareOperator returns [BinaryOperator op = null; ]
    :   EQUAL!        { op = BinaryOperator.EQUAL; }
    |   NOT_EQUAL!    { op = BinaryOperator.NOT_EQUAL; }
    |   LT_EQUAL!     { op = BinaryOperator.LESS_EQUAL; }
    |   GT_EQUAL!     { op = BinaryOperator.GREATER_EQUAL; }
    |   LT!           { op = BinaryOperator.LESS; }
    |   GT!           { op = BinaryOperator.GREATER; }
    ;

connectOperator returns [BinaryOperator op = null; ]
    :   LSHIFT!       { op = BinaryOperator.CONNECTED_TO_FROM; }
    ;


simpleExpression returns [SimpleExpression expr = null] 
    {
        UnaryOperator op = null;
    }
    :   expr = parExpression
    |   op = unaryOperator expr = simpleExpression
        {
            UnaryExpression n = factory_.createUnaryExpression();
            n.setOperator( op );
            n.setExpression( expr );
            expr = n;
        }
    |   expr=portExpression
    |   expr=literalExpression
    |   expr=activeExpression
    ;
    
parExpression returns [UnaryExpression expr = null]
    { Expression e = null; }
    :    LPAREN! e = expression RPAREN!
        {
            expr = factory_.createUnaryExpression();
            expr.setOperator( UnaryOperator.NONE );
            expr.setExpression( e );
        }
    ;

unaryOperator returns [UnaryOperator op = null]
    :    "always" { op = UnaryOperator.ALWAYS; }
    |    "once"   { op = UnaryOperator.ONCE; }
    |    "prev"   { op = UnaryOperator.PREVIOUS; }
    |    NOT      { op = UnaryOperator.NOT; }
    ;

portExpression returns [PortExpression expr = null]
    :   i1: identifier
        ( DOT i2: identifier)?
        {
            String portName = #i1.getText();
            Component c = null;
            if( #i2 != null ) {
                portName = #i2.getText();
                c = helper_.getApplicationComponent( #i1.getText() );
            }
            expr = factory_.createPortExpression( );
            Port port = c == null
                ? (
                    application_ == null
                        ? helper_.getComponentPort( component_, portName )
                        : helper_.getApplicationPort( portName )
                )
                : helper_.getComponentPort( c.getType(), portName );
            if( port == null )
                reportError(#i1.getLine(), #i1.getColumn(), "Port \"" + portName + "\" not found." );
            expr.setPort( port );
            if( c != null ) expr.setComponent( c );
         }
    ;

literalExpression returns [LiteralExpression expr = null]
    :   l: literal
        {
            expr = factory_.createLiteralExpression();
            expr.setValue( #l.getText() );
        }
    ;
    
activeExpression returns [ActiveExpression e = null]
    : "active" LPAREN i:identifier RPAREN
    {
            e = factory_.createActiveExpression();
            String compName = #i.getText();
            Component comp = helper_.getApplicationComponent( compName );
            if( comp == null ) reportError(#i.getLine(), #i.getColumn(), "In \"active\" expression: component " + compName + " not found." );
            e.setComponent( comp );
            e.setActive( true );
    }
    ;

never_clause returns [List<SafetyClause> clauses = new ArrayList<SafetyClause>() ]
            { Expression e; }
//    :   "never"^ LCURLY! (temporalExpression SEMI!)* RCURLY
    :   "never"^ LCURLY!
        (
            e = expression SEMI!     
            {
                NeverClause c = factory_.createNeverClause();
                c.setExpression( e );
                clauses.add( c );
            }
        )*
        RCURLY
    ;

always_clause returns [List<SafetyClause> clauses = new ArrayList<SafetyClause>() ]
            { Expression e; }
//    :   "always"^ LCURLY! (temporalExpression SEMI!)* RCURLY
    :   "always"^ LCURLY! (e=expression SEMI! 
    {
        AlwaysClause c = factory_.createAlwaysClause();
        c.setExpression( e );
        clauses.add( c );
    }
    )* RCURLY
    ;


literal
    :   integer_literal
    |   floating_pt_literal
    ;

integer_literal
    :   INT
    |   OCTAL
    |   HEX
    ;

floating_pt_literal
    :   f:FLOAT
    ;

identifier
    :   IDENT^
    ;

/* IDL LEXICAL RULES  */
class ADLVLexer extends Lexer;
options {
    exportVocab=ADLV;
    charVocabulary='\u0000'..'\uFFFE';
    k=4;
}

{
    private Vector errors = new Vector();
    private Vector warnings = new Vector();

    /**@return The number of errors found in the lexer*/
    public int errorNumber(){
        return errors.size();
    }

    /**@return The number of warnings found in the lexer*/
    public int warningNumber(){
        return warnings.size();
    }

    /**@return All errors messages found in the lexer*/
    public String [] errors () {
        String [] ret = new String [errors.size()];
        for (int i = 0; i < ret.length; ++i)
            ret[i] = (String)errors.elementAt(i);
        return ret;
    }

    /**@return All warning messages found in the lexer*/
    public String [] warnings () {
        String [] ret = new String [warnings.size()];
        for (int i = 0; i < ret.length; ++i)
            ret[i] = (String)warnings.elementAt(i);
        return ret;
    }

    public void reportError(RecognitionException ex) {
        errors.addElement(ex.toString());
    }

    public void reportError(String s) {
        errors.addElement(s);
    }

    public void reportWarning(String s) {
        warnings.addElement(s);
    }
}

SEMI
options {
  paraphrase = ";";
}
    :    ';'
    ;

QUESTION
options {
  paraphrase = "?";
}
    :    '?'
    ;

IMPLIES
options {
  paraphrase = "=>";
}
    :    "=>"
    ;

LPAREN
options {
  paraphrase = "(";
}
    :    '('
    ;

RPAREN
options {
  paraphrase = ")";
}
    :    ')'
    ;

LBRACK
options {
  paraphrase = "[";
}
    :    '['
    ;

RBRACK
options {
  paraphrase = "]";
}
    :    ']'
    ;

LCURLY
options {
  paraphrase = "{";
}
    :    '{'
    ;

RCURLY
options {
  paraphrase = "}";
}
    :    '}'
    ;

OR
options {
  paraphrase = "|";
}
    :    '|'
    ;

LOGICAL_OR
options {
  paraphrase = "||";
}
    :    "||"
    ;

XOR
options {
  paraphrase = "^";
}
    :    '^'
    ;

AND
options {
  paraphrase = "&";
}
    :    '&'
    ;

LOGICAL_AND
options {
  paraphrase = "&&";
}
    :    "&&"
    ;

COLON
options {
  paraphrase = ":";
}
    :    ':'
    ;

COMMA
options {
  paraphrase = ",";
}
    :    ','
    ;

DOT
options {
  paraphrase = ".";
}
    :    '.'
    ;

ASSIGN
options {
  paraphrase = "=";
}
    :    '='
    ;

NOT
options {
  paraphrase = "!";
}
    :    '!'
    ;

EQUAL
options {
  paraphrase = "==";
}
    :    "=="
    ;

NOT_EQUAL
options {
  paraphrase = "!=";
}
    :    "!="
    ;

LT
options {
  paraphrase = "<";
}
    :    '<'
    ;

LT_EQUAL
options {
  paraphrase = "<=";
}
    :    "<="
    ;

LSHIFT
options {
  paraphrase = "<<";
}
    : "<<"
    ;

GT
options {
  paraphrase = ">";
}
    :    '>'
    ;

GT_EQUAL
options {
  paraphrase = ">=";
}
    :    ">="
    ;

RSHIFT
options {
  paraphrase = ">>";
}
    : ">>"
    ;

DIV
options {
  paraphrase = "/";
}
    :    '/'
    ;

PLUS
options {
  paraphrase = "+";
}
    :    '+'
    ;

MINUS
options {
  paraphrase = "-";
}
    :    '-'
    ;

TILDE
options {
  paraphrase = "~";
}
    :    '~'
    ;

STAR
options {
  paraphrase = "*";
}
    :    '*'
    ;

MOD
options {
  paraphrase = "%";
}
    :    '%'
    ;


SCOPEOP
options {
  paraphrase = "::";
}
    :      "::"
    ;

WS
options {
  paraphrase = "white space";
}
    :    (' '
    |    '\t'
    |    '\n'  { newline(); }
    |    '\r')
        { $setType(Token.SKIP); }
    ;


PREPROC_DIRECTIVE
options {
  paraphrase = "a preprocessor directive";
}

    :
    '#'!
    (~'\n')* '\n'!
    { $setType(Token.SKIP); newline(); }
    ;


SL_COMMENT
options {
  paraphrase = "a comment";
}

    :
    "//"!
    (~'\n')* '\n'
    { $setType(Token.SKIP); newline(); }
    ;

ML_COMMENT
options {
  paraphrase = "a comment";
}
    :
    "/*"!
    (
            '\n' { newline(); }
        |    ('*')+
            (    '\n' { newline(); }
            |    ~('*' | '/' | '\n')
            )
        |    ~('*' | '\n')
    )*
    "*/"!
    { $setType(Token.SKIP);  }
    ;

CHAR_LITERAL
options {
  paraphrase = "a character literal";
}
    :
    '\''!
    ( ESC | ~'\'' )
    '\''!
    ;

WIDE_CHAR_LITERAL
options {
  paraphrase = "a wide character literal";
}
    : 'L'! CHAR_LITERAL
    ;

STRING_LITERAL
options {
  paraphrase = "a string literal";
}
    :
    '"'!
    (ESC|~'"')*
    '"'!
    ;


WIDE_STRING_LITERAL
options {
  paraphrase = "a wide string literal";
}
    :
    'L'! STRING_LITERAL
    ;

protected
ESC
options {
  paraphrase = "an escape sequence";
}
    :    '\\'!
        (    'n'        {$setText("\n");}
        |    't'        {$setText("\t");}
        |    'v'        {$setText("\013");}
        |    'b'        {$setText("\b");}
        |    'r'        {$setText("\r");}
        |    'f'        {$setText("\r");}
        |    'a'          {$setText("\007");}
        |    '\\'        {$setText("\\");}
        |    '?'         {$setText("?");}
        |    '\''        {$setText("'");}
        |    '"'        {$setText("\"");}
        |    OCTDIGIT
            (options {greedy=true;}:OCTDIGIT
              (options {greedy=true;}:OCTDIGIT)?
            )?
            {char realc = (char) Integer.valueOf($getText, 8).intValue(); $setText(realc);}
        |       'x'! HEXDIGIT
            (options {greedy=true;}:HEXDIGIT)?
            {char realc = (char) Integer.valueOf($getText, 16).intValue(); $setText(realc);}
        |    'u'!
            HEXDIGIT
            (options {greedy=true;}:HEXDIGIT
              (options {greedy=true;}:HEXDIGIT
                (options {greedy=true;}:HEXDIGIT)?
              )?
            )?
            {char realc = (char) Integer.valueOf($getText, 16).intValue(); $setText(realc);}
        )
    ;

protected
VOCAB
options {
  paraphrase = "an escaped character value";
}
    :    '\3'..'\377'
    ;

protected
DIGIT
options {
  paraphrase = "a digit";
}
    :    '0'..'9'
    ;

protected
NONZERODIGIT
options {
  paraphrase = "a non-zero digit";
}
    :    '1'..'9'
    ;

protected
OCTDIGIT
options {
  paraphrase = "an octal digit";
}
    :    '0'..'7'
    ;

protected
HEXDIGIT
options {
  paraphrase = "a hexadecimal digit";
}
    :    ('0'..'9' | 'a'..'f' | 'A'..'F')
    ;

HEX
options {
  paraphrase = "a hexadecimal value value";
}

    :    ("0x" | "0X") (HEXDIGIT)+
    ;

INT
options {
  paraphrase = "an integer value";
}
    :    NONZERODIGIT (DIGIT)*                  // base-10
         (  '.' (DIGIT)*
         ( (('e' | 'E') ('+' | '-')? (DIGIT)+)    {$setType(FLOAT);}
         |                    {$setType(FLOAT);}
         )
         |   ('e' | 'E') ('+' | '-')? (DIGIT)+       {$setType(FLOAT);}
         )?
    ;

OCTAL
options {
  paraphrase = "an octal value";
}
    :    '0'
         ( (DIGIT)+
         | FLOAT                    {$setType(FLOAT);}
         |                        {$setType(INT);}
         )
    ;


FLOAT
options {
  paraphrase = "a floating point value";
}

    :    '.' (DIGIT)+
         ( ('e' | 'E') ('+' | '-')? (DIGIT)+
         )?
    ;

IDENT
options {
  paraphrase = "an identifer";
  testLiterals = true;
}

    :   ('a'..'z'|'A'..'Z') ('a'..'z'|'A'..'Z'|'_'|'0'..'9')*
    ;

ESCAPED_IDENT
options {
  paraphrase = "an escaped identifer";
  testLiterals = false;            // redundant, but explicit is good.
}
    // NOTE: Adding a ! to the '_' doesn't seem to work,
    //       so we adjust _begin manually.

    :   '_' ('a'..'z'|'A'..'Z') ('a'..'z'|'A'..'Z'|'_'|'0'..'9')*
                            {_begin++;$setType(IDENT);}
    ;

